package com.deviceatlas.examples;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import java.util.HashMap;
import java.util.Map;
import java.util.stream.Collectors;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Demand Side Platform (DSP) CLI example of how the OpenRTB SUA object (OpenRTB 2.6 specification SUA (Structured User Agent))
 * can be read and parsed to standard HTTP header format
 *
 * @author DeviceAtlas Limited (c)
 */
public class DSPExample {

    private final Gson gson;
    private static final  String USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36";

    /**
     * Example of the SUA object (OpenRTB-2-6 Documentation - 3.2.29 Object: UserAgent)
     * which is part of the Device Object (OpenRTB-2-6 Documentation - 3.2.18 Object: Device)
     */
    private static final String SUA_OBJECT = "{\"sua\":{\"browsers\":[{\"brand\":\"Google Chrome\",\"version\":[\"107\",\"0\",\"5304\",\"87\"]},{\"brand\":\"Chromium\",\"version\":[\"107\",\"0\",\"5304\",\"87\"]},{\"brand\":\"Not=\\\\\\\"A?Brand\",\"version\":[\"24\",\"0\",\"0\",\"0\"]}],\"platform\":{\"brand\":\"macOS\",\"version\":[\"13\",\"0\",\"0\"]},\"mobile\":\"1\",\"architecture\":\"x86\",\"bitness\":\"64\",\"model\":\"\"}}";

    public DSPExample() {
        gson = new GsonBuilder()
                .setPrettyPrinting()
                .disableHtmlEscaping()
                .create();
    }

    public static void main (String [] args){
        DSPExample dspExample = new DSPExample();
        dspExample.parserSuaObject();
    }

    /**
     * Parse the SUA Object to a headers Map
     */
    private void parserSuaObject() {
        System.out.println("SUA JSON Object:%n %s".formatted(gson.toJson(JsonParser.parseString(SUA_OBJECT))));
        System.out.println("-------------------------------------------------------------");
        JsonObject suaObject = this.gson.fromJson(SUA_OBJECT, JsonObject.class);
        suaObject = suaObject.getAsJsonObject("sua");

        String brands = parseBrowsersToBrands(suaObject);
        String brandsWithSignificantVersion = parseBrowsersToBrandsWithSignificantVersionNumber(suaObject);
        JsonObject platformObject = suaObject.getAsJsonObject("platform");

        // Filling the Client Hints header with the SUA Object values
        Map<String, String> headers = new HashMap<>();
        headers.put("User-agent", USER_AGENT);
        headers.put("Sec-Ch-Ua", brandsWithSignificantVersion);
        headers.put("Sec-Ch-Ua-Full-Version-List", brands);
        headers.put("Sec-Ch-Ua-Platform", addDoubleQuotes(getBrand(platformObject)));
        headers.put("Sec-Ch-Ua-Platform-Version", addDoubleQuotes(parseVersionArrayToString(platformObject)));
        headers.put("Sec-Ch-Ua-Model", addDoubleQuotes(suaObject.get("model").getAsString()));
        headers.put("Sec-Ch-Ua-Arch", addDoubleQuotes(suaObject.get("architecture").getAsString()));
        headers.put("Sec-Ch-Ua-Bitness", addDoubleQuotes(suaObject.get("bitness").getAsString()));

        /**
         * The Sec-Ch-Ua-Mobile header uses ?1 for true values and ?0 for false values
         */
        if (suaObject.get("mobile").getAsString().equals("1")){
            headers.put("Sec-Ch-Ua-Mobile", "?1");
        }else{
            headers.put("Sec-Ch-Ua-Mobile", "?0");
        }

        System.out.println("Headers Parsed:%n%s".formatted(getHeadersFormatted(headers)));
        System.out.println("-------------------------------------------------------------");
    }

    /**
     * Some Client Hints headers have their value enclosed in double quotes
     * @param value
     * @return
     */
    private String addDoubleQuotes(String value) {
        if (value.equals("")){
            return value;
        }

        return "\\\"%s\\\"".formatted(value);
    }

    /**
     * Get the headers as a String formatted
     * @param headers
     * @return
     */
    private String getHeadersFormatted(Map<String, String> headers) {
        return headers.entrySet()
                .stream()
                .map(e -> e.getKey() + "=\"" + e.getValue() + "\"")
                .collect(Collectors.joining("\n"));
    }

    /**
     * Parse the browsers object to the brand string
     * @param suaObject
     * @return
     */
    private String parseBrowsersToBrands(JsonObject suaObject) {
        StringBuilder brands = new StringBuilder();
        JsonArray browsersArray = suaObject.getAsJsonArray("browsers");

        for (int i = 0; i <browsersArray.size(); i++) {
            JsonObject browserObject = browsersArray.get(i).getAsJsonObject();

            String brand = getBrand(browserObject);
            String versions = parseVersionArrayToString(browserObject);

            if (!brands.isEmpty()){
                brands.append(", ");
            }

            brands.append("\\\"%s\";v=\\\"%s\\\"".formatted(brand, versions));
        }
        return brands.toString();
    }

     /**
     * Parse the browsers object to the brand string
     * containg the significant version number value
     * @param suaObject
     * @return
     */
    private String parseBrowsersToBrandsWithSignificantVersionNumber(JsonObject suaObject) {
        StringBuilder brands = new StringBuilder();
        JsonArray browsersArray = suaObject.getAsJsonArray("browsers");

        for (int i = 0; i <browsersArray.size(); i++) {
            JsonObject browserObject = browsersArray.get(i).getAsJsonObject();

            String brand = getBrand(browserObject);
            String significantVersionNumber = parseVersionArrayToSignificantVersioNumber(browserObject);

            if (!brands.isEmpty()){
                brands.append(", ");
            }

            brands.append("\\\"%s\";v=\\\"%s\\\"".formatted(brand, significantVersionNumber));
        }
        return brands.toString();
    }

    /**
     * Parse the versions array ["102", "0", "5005", "115"] to a string
     * containing the values contacted separated by a dot E.G 102.0.5005.115
     * @param jsonObject
     * @return
     */
    private String parseVersionArrayToString(JsonObject jsonObject) {
        JsonArray versionJsonArray = jsonObject.get("version").getAsJsonArray();

        StringBuilder versions = new StringBuilder();
        for (JsonElement version : versionJsonArray.asList()){
            if (!versions.isEmpty()){
                versions.append(".");
            }

            versions.append(version.getAsString());

        }
        return versions.toString();
    }

    /**
     * Parse the versions array ["102", "0", "5005", "115"] to collect
     * the significant version number
     * @param jsonObject
     * @return
     */
    private String parseVersionArrayToSignificantVersioNumber(JsonObject jsonObject) {
        JsonArray versionJsonArray = jsonObject.get("version").getAsJsonArray();
        return versionJsonArray.asList().get(0).getAsString();
    }

    /**
     * Get the Brand present in the JSON object
     * @param jsonObject
     * @return
     */
    private String getBrand(JsonObject jsonObject) {
        return jsonObject.get("brand").getAsString();
    }
}